import os
from dataclasses import dataclass
from typing import Any
from typing import NamedTuple
from typing import Protocol

import requests
from dapla_auth_client import AuthClient

from ssb_pubmd.adapters.content_parser import Content
from ssb_pubmd.config import Config


class PublishClientError(Exception): ...

class HttpClient(Protocol):
    def post(
        self, url: str, headers: dict[str, str], payload: dict[str, Any]
    ) -> dict[str, str]: ...

class RequestsHttpClient:
    def post(
        self, url: str, headers: dict[str, str], payload: dict[str, Any]
    ) -> dict[str, str]:
        try:
            response = requests.post(
                url,
                headers=headers,
                json=payload,
            )
            body = response.json()
        except Exception as err:
            raise PublishClientError("Sync failed. Could not parse response.") from err
        if not response.ok:
            raise PublishClientError(
                f"Sync failed. Response message: {body.get('msg', 'no message')}"
            )
        return body  # type: ignore

class TokenClient(Protocol):
    def get_token(self) -> str: ...


class LocalTokenClient:
    def get_token(self) -> str:
        return os.environ.get("OIDC_TOKEN", "")


class DaplaTokenClient:
    def get_token(self) -> str:
        return AuthClient.fetch_personal_token(audiences=["ssbno"])

class Response(NamedTuple):
    publish_path: str
    publish_id: str
    publish_url: str
    publish_html: str

class PublishClient(Protocol):
    http_client: HttpClient

    def send_content(self, content: Content) -> Response: ...

DEFAULT_HTTP_CLIENT = RequestsHttpClient()
DEFULT_TOKEN_CLIENT = LocalTokenClient()
@dataclass
class MimirPublishClient:
    base_url: str
    endpoint: str
    preview_base_path: str
    http_client: HttpClient = DEFAULT_HTTP_CLIENT
    token_client: TokenClient = DEFULT_TOKEN_CLIENT

    def _create_headers(self) -> dict[str, str]:
        return {
            "Authorization": f"Bearer {self.token_client.get_token()}",
            "Content-Type": "application/json",
        }

    def send_content(self, content: Content) -> Response:
        headers = self._create_headers()
        response_body = self.http_client.post(
            url=f"{self.base_url}{self.endpoint}",
            headers=headers,
            payload=content.serialize(),
        )

        id_ = response_body.get("_id")
        path = response_body.get("_path")

        if path is None or id_ is None:
            raise PublishClientError("Sync failed. Could not parse response body.")

        macro_type = (
            content.content_type
            if content.content_type in ["highchart", "factBox"]
            else None
        )
        if id_ is not None and macro_type is not None:
            html = f"<p>[ {macro_type} {content.content_type}=&quot;{id_}&quot; /]</p>"
        else:
            html = ""

        return Response(
            publish_path=path,
            publish_id=id_,
            publish_url=self.base_url + self.preview_base_path + path,
            publish_html=html,
        )


def get_publish_client(config: Config) -> PublishClient:
    return MimirPublishClient(
        base_url=config.publish_base_url,
        endpoint="/webapp/pubmd",
        preview_base_path=config.publish_admin_path + "/site/preview/default/draft",
        token_client=DaplaTokenClient()
        if config.use_dapla_token_client
        else DEFULT_TOKEN_CLIENT,
    )



import json
import subprocess
from collections.abc import Iterator
from typing import Any
from typing import NamedTuple
from typing import Protocol
from typing import TypedDict

import pandocfilters as pf  # type: ignore


class Element(NamedTuple):
    id: str
    inner_html: str | None


class DocumentProcessor(Protocol):
    def load(self, raw_content: str) -> None: ...
    def extract_metadata(self, target_key: str) -> dict[str, Any]: ...
    def extract_elements(self, target_class: str) -> Iterator[Element]: ...
    def replace_element(self, id_: str, new_html: str) -> None: ...
    def extract_html(self) -> str: ...



class PandocElement(TypedDict):
    t: str
    c: Any


PandocDocument = TypedDict(
    "PandocDocument",
    {
        "pandoc-api-version": list[int],
        "meta": dict[str, Any],
        "blocks": list[PandocElement],
    },
)


class PandocDocumentProcessor:
    """
    Processor for a pandoc document, i.e. the JSON-serialized pandoc AST of a document.

    Example pandoc AST with exactly one div:

    ```json
    {
        "pandoc-api-version": [1, 23, 1],
        "meta": {},
        "blocks": [
        {
            "t": "Div",
            "c": [
            ["my-highchart", ["ssb"], [["title", "My highchart"]]],
            []
            ]
        }
        ]
    }
    ```
    Html equivalent:
    ```html
    <div id="my-highchart" class="ssb" title="My highchart">
    </div>
    ```
    References:
    - Studying the result of command `pandoc FILE -t json`, where FILE is a minimal example document (e.g. Markdown or html).
    - https://github.com/jgm/pandocfilters has some examples of how to work with the format.
    - Note: no formal specification exists.
    """

    document: PandocDocument
    _element_index: dict[str, int]

    def load(self, raw_content: str) -> None:
        self.document: PandocDocument = json.loads(raw_content)
        self._element_index = {}

    def extract_metadata(self, target_key: str) -> dict[str, Any]:
        def meta_to_dict(meta: Any) -> Any:
            t, c = meta.get("t"), meta.get("c")
            if t == "MetaMap":
                return {k: meta_to_dict(v) for k, v in c.items()}
            elif t == "MetaList":
                return [meta_to_dict(v) for v in c]
            else:
                return pf.stringify(c)

        return meta_to_dict(self.document["meta"][target_key])  # type: ignore

    def extract_html(self) -> str:
        return self._document_to_html(self.document)

    def extract_elements(self, target_class: str) -> Iterator[Element]:
        self._element_index = self._generate_element_index(target_class)

        for id_, i in self._element_index.items():
            element = self.document["blocks"][i]
            inner_blocks: list[PandocElement] = element["c"][1]
            inner_html = self._blocks_to_html(inner_blocks) if inner_blocks else None
            yield Element(id_, inner_html)

    def replace_element(self, id_: str, new_html: str) -> None:
        i = self._element_index[id_]
        self.document["blocks"][i] = {
            "t": "RawBlock",
            "c": ["html", new_html],
        }

    def _generate_element_index(self, target_class: str) -> dict[str, int]:
        index = {}
        for i, element in enumerate(self.document["blocks"]):
            if element["t"] != "Div":
                continue

            id_: str = element["c"][0][0]
            if not id_:
                continue

            classes: list[str] = element["c"][0][1]
            if target_class not in classes:
                continue

            index[id_] = i

        return index

    @classmethod
    def _blocks_to_html(cls, blocks: list[PandocElement]) -> str:
        document: PandocDocument = {
            "pandoc-api-version": [1, 23, 1],
            "meta": {},
            "blocks": blocks,
        }
        return cls._document_to_html(document)

    @classmethod
    def _document_to_html(cls, document: PandocDocument) -> str:
        result = subprocess.run(
            ["pandoc", "-f", "json", "-t", "html"],
            input=json.dumps(document),
            text=True,
            capture_output=True,
            check=True,
        )
        html = result.stdout
        return html


import sys

from ssb_pubmd.cli import run_cli
from ssb_pubmd.config import get_config


def main() -> None:
    config = get_config()
    run_cli(sys.argv, config)


if __name__ == "__main__":
    main()



from collections.abc import Mapping
from dataclasses import asdict
from dataclasses import dataclass
from typing import Any
from typing import Literal
from typing import Protocol

import nh3


@dataclass
class Content:
    title: str
    content_type: str
    publish_folder: str | None = None
    publish_id: str | None = None

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)

    def serialize(self) -> dict[str, Any]:
        raise NotImplementedError()

class ContentParser(Protocol):
    def parse(self, metadata: Mapping[str, Any], html: str | None) -> Content: ...


@dataclass
class MimirContent(Content):
    def is_publishable(self) -> bool:
        if self.title == "":
            return False
        if self.publish_id is None and self.publish_folder is None:
            return False
        return True

    def serialize(self) -> dict[str, Any]:
        if not self.is_publishable():
            raise Exception()
        s: dict[str, Any] = {
            "contentType": "mimir:" + self.content_type,
            "displayName": self.title,
            "parentPath": self.publish_folder,
            "data": {},
        }
        if self.publish_id is not None:
            s["_id"] = self.publish_id
        return s
    

@dataclass
class Author:
    name: str
    email: str

@dataclass
class Article(MimirContent):
    content_type: str = "article"
    authors: list[Author] | None = None
    ingress: str = ""
    html_text: str = ""

    def serialize(self) -> dict[str, Any]:
        s = super().serialize()
        if self.authors:
            s["data"]["authorItemSet"] = [asdict(author) for author in self.authors]
        s["data"]["ingress"] = self.ingress
        s["data"]["articleText"] = self.html_text
        return s


GraphType = Literal["line", "pie", "column", "bar", "area", "barNegative"]


@dataclass
class Highchart(MimirContent):
    content_type: str = "highchart"
    graph_type: GraphType = "line"
    html_table: str | None = None
    tbml: str | None = None
    xlabel: str = "x"
    ylabel: str = "y"

    def serialize(self) -> dict[str, Any]:
        s = super().serialize()

        if self.html_table is not None:
            s["data"]["htmlTable"] = self.html_table
        elif self.tbml is not None:
            s["data"]["dataSource"] = {
                "_selected": "tbprocessor",
                "tbprocessor": {"urlOrId": self.tbml},
            }

        s["data"]["xAxisTitle"] = self.xlabel
        s["data"]["yAxisTitle"] = self.ylabel

        return s


@dataclass
class FactBox(MimirContent):
    content_type: str = "factBox"
    display_type: Literal["default", "sneakPeek", "aiIcon"] = "default"
    html_text: str = ""

    def serialize(self) -> dict[str, Any]:
        s = super().serialize()
        s["data"]["expansionBoxType"] = self.display_type
        s["data"]["text"] = self.html_text
        return s


BASIC_HTML_TAGS = {
    "p",
    "br",
    "strong",
    "em",
    "b",
    "i",
    "ul",
    "ol",
    "li",
    "blockquote",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "a",
}


class MimirContentParser:
    def parse(self, metadata: Mapping[str, Any], html: str | None) -> Content:
        match metadata.get("content_type"):
            case "article":
                return self._parse_article(metadata, html)
            case "factBox":
                return self._parse_factbox(metadata, html)
            case "highchart":
                return self._parse_highchart(metadata, html)
            case _:
                return MimirContent(**metadata)

    def serialize(self, content: Content) -> dict[str, Any]:
        if isinstance(content, MimirContent):
            return content.serialize()
        else:
            raise Exception()

    @classmethod
    def _parse_article(cls, metadata: Mapping[str, Any], html: str | None) -> Article:
        article = Article(
            title=metadata["title"],
            publish_folder="/ssb" + metadata["path"],
            publish_id=metadata.get("publish_id"),
            authors=[Author(**data) for data in metadata.get("authors", [])],
            ingress=metadata.get("ingress", ""),
        )
        if html is not None:
            allowed_html_tags = BASIC_HTML_TAGS
            html_text = nh3.clean(html, tags=allowed_html_tags)
            article.html_text = html_text
        return article

    @classmethod
    def _parse_factbox(cls, metadata: Mapping[str, Any], html: str | None) -> FactBox:
        factbox = FactBox(**metadata)
        if html is not None:
            allowed_html_tags = BASIC_HTML_TAGS - {"h2"}
            html_text = nh3.clean(html, tags=allowed_html_tags)
            factbox.html_text = html_text
        return factbox

    @classmethod
    def _parse_highchart(
        cls, metadata: Mapping[str, Any], html: str | None
    ) -> Highchart:
        highchart = Highchart(**metadata)
        if html is not None:
            allowed_html_tags = {"table", "tbody", "tr", "td"}
            html_table = nh3.clean(html, tags=allowed_html_tags)
            highchart.html_table = html_table
        return highchart


from pathlib import Path
from typing import Literal

import narwhals as nw
from narwhals.typing import IntoDataFrame

from ssb_pubmd.adapters.content_parser import Content
from ssb_pubmd.adapters.content_parser import ContentParser
from ssb_pubmd.adapters.content_parser import MimirContentParser
from ssb_pubmd.adapters.storage import LocalFileStorage
from ssb_pubmd.adapters.storage import Storage
from ssb_pubmd.domain.document_publisher import USER_KEY_PREFIX

STORAGE: Storage = LocalFileStorage(project_folder=Path.cwd())
CONTENT_PARSER: ContentParser = MimirContentParser()

class NotebookClientError(Exception): ...

def configure_factbox(
    key: str,
    title: str,
    display_type: Literal["default", "sneakPeek", "aiIcon"] = "default",
) -> None:
    """Oppretter en faktaboks og printer en Markdown-snippet som kan limes inn i artikkelen (på en ny linje).

    :param key: En unik nøkkel for innholdet.
    :param title: Tittelen til faktaboksen.
    :param display_type: Visning av faktaboksen.

        Alternativer:

        * "default": Bare tittel (standard)
        * "sneakPeek": Tittel og litt av forklaringsteksten
        * "aiIcon": Tittel og litt av forklaringsteksten + KI-ikon
    """
    metadata = {
        "content_type": "factBox",
        "title": title,
        "display_type": display_type,
    }

    content = CONTENT_PARSER.parse(
        metadata=metadata,
        html=None,
    )
    _store_user_content(user_key=key, content=content)

    md = _get_markdown_snippet(
        key,
        placeholder_text="Faktaboksens tekst skrives her med Markdown-syntaks.\n\n### Underoverskrift\n\nNy paragraf.",
    )
    print(md)


def create_highchart(
    key: str,
    title: str,
    dataframe: IntoDataFrame | None = None,
    tbml: str | None = None,
    graph_type: Literal["line", "pie", "column", "bar", "area", "barNegative"] = "line",
    xlabel: str = "x",
    ylabel: str = "y"
) -> None:
    """Oppretter et highchart og printer en Markdown-snippet som kan limes inn i artikkelen (på en ny linje).

    Som datakilde er det nødvendig å spesifisere enten `dataframe` eller `tbml`.

    :param key: En unik nøkkel for innholdet.
    :param title: Tittelen til highchartet.
    :param dataframe: En pandas, Polars eller PyArrow dataframe.
    :param tbml: URL eller TBML-id.
    :param graph_type: Graftype.

        Alternativer:

        * "line": Linje (standard)
        * "pie": Kake
        * "column": Stolpe
        * "bar": Liggende stolpe
        * "area": Areal
        * "barNegative": Pyramide

    :param xlabel: X-akse, tittel.
    :param ylabel: Y-akse, tittel.
    """
    if dataframe is None and tbml is None:
        raise NotebookClientError("Either 'dataframe' or 'tbml' must be specified.")
    metadata = {
        "content_type": "highchart",
        "title": title,
        "graph_type": graph_type,
        "xlabel": xlabel,
        "ylabel": ylabel
    }
    if tbml is not None:
        metadata["tbml"] = tbml
    html = _dataframe_to_html_table(dataframe) if dataframe is not None else None
    content = CONTENT_PARSER.parse(metadata, html)
    _store_user_content(user_key=key, content=content)

    md = _get_markdown_snippet(key)
    print(md)


def _store_user_content(user_key: str, content: Content) -> None:
    key = USER_KEY_PREFIX + str(user_key)
    STORAGE.update(key, content.to_dict())


def _dataframe_to_html_table(dataframe: IntoDataFrame) -> str:
    df = nw.from_native(dataframe)
    html = "<table><tbody>\n"

    html += "<tr>\n"
    for name in df.columns:
        html += f"    <td>{name}</td>\n"
    html += "</tr>\n"

    for row in df.iter_rows():
        html += "  <tr>\n"
        for value in row:
            html += f"    <td>{value}</td>\n"
        html += "  </tr>\n"

    html += "</tbody></table>"

    return html


def _get_markdown_snippet(key: str, placeholder_text: str | None = None) -> str:
    div_config = f"{{ #{key} .ssb }}"
    div_content = f"\n{placeholder_text}\n\n" if placeholder_text is not None else ""
    return f"::: {div_config}\n{div_content}:::"


from ssb_pubmd.adapters.content_parser import ContentParser
from ssb_pubmd.adapters.document_processor import DocumentProcessor
from ssb_pubmd.adapters.publish_client import PublishClient
from ssb_pubmd.adapters.storage import Storage

USER_KEY_PREFIX = "user:"
DOCUMENT_KEY = "app:document"

def sync_document(
    raw_document_content: str,
    document_processor: DocumentProcessor,
    content_parser: ContentParser,
    storage: Storage,
    publish_client: PublishClient,
) -> str:
    document_processor.load(raw_document_content)

    document_metadata = document_processor.extract_metadata(target_key="ssb")
    document_metadata["content_type"] = "article"

    document_publish_path = storage.get(DOCUMENT_KEY).get("publish_path")
    if not document_publish_path:
        content = content_parser.parse(metadata=document_metadata, html=None)
        response = publish_client.send_content(content)
        storage.update(
            DOCUMENT_KEY,
            {"publish_id": response.publish_id, "publish_path": response.publish_path},
        )
        document_publish_path = response.publish_path

    document_elements = document_processor.extract_elements(target_class="ssb")
    for id_, html in document_elements:
        key = USER_KEY_PREFIX + id_
        metadata = storage.get(key) | {"publish_folder": document_publish_path}
        component = content_parser.parse(metadata, html)
        response = publish_client.send_content(component)
        storage.update(key, {"publish_id": response.publish_id})
        document_processor.replace_element(id_, response.publish_html)

    article_metadata = document_metadata | {
        "publish_id": storage.get(DOCUMENT_KEY).get("publish_id")
    }
    html = document_processor.extract_html()
    article = content_parser.parse(metadata=article_metadata, html=html)
    response = publish_client.send_content(article)
    return response.publish_url



import json
from collections.abc import Mapping
from pathlib import Path
from typing import Any
from typing import Protocol


class Storage(Protocol):
    def update(self, key: str, data: Mapping[str, Any]) -> None: ...
    def get(self, key: str) -> dict[str, Any]: ...

class LocalFileStorage:
    path: Path

    def __init__(self, project_folder: Path) -> None:
        self.path = project_folder / ".ssbno.json"
        if not self.path.exists():
            with self.path.open("w") as f:
                json.dump({}, f)

    def _load(self) -> dict[str, dict[str, Any]]:
        with self.path.open() as f:
            return json.load(f)  # type: ignore

    def _save(self, data: dict[str, dict[str, Any]]) -> None:
        with self.path.open("w") as f:
            json.dump(data, f, indent=2)

    def update(self, key: str, data: Mapping[str, Any]) -> None:
        store = self._load()

        current = store.get(key, {})
        for field, value in data.items():
            if value is not None:
                current[field] = value

        store[key] = current
        self._save(store)

    def get(self, key: str) -> dict[str, Any]:
        store = self._load()
        return store.get(key, {}).copy()


import subprocess
import sys
from importlib import resources
from pathlib import Path

from jinja2 import Template
from watchfiles import watch

from ssb_pubmd import templates
from ssb_pubmd.adapters.content_parser import MimirContentParser
from ssb_pubmd.adapters.document_processor import PandocDocumentProcessor
from ssb_pubmd.adapters.publish_client import PublishClient
from ssb_pubmd.adapters.publish_client import get_publish_client
from ssb_pubmd.adapters.storage import LocalFileStorage
from ssb_pubmd.config import Config
from ssb_pubmd.domain.document_publisher import sync_document


def run_cli(system_arguments: list[str], config: Config) -> None:
    match system_arguments:
        case [_, "create", filename]:
            _create_template(filename, config)
        case [_, "preview", filename]:
            _preview(filename, config)
        case _:
            print("Usage:")
            print("  ssb-pubmd create FILENAME")
            print("  ssb-pubmd preview FILENAME")
            sys.exit(1)


def _create_template(filename: str, config: Config) -> None:
    destination = Path.cwd() / filename
    if destination.suffix != ".qmd":
        print("The destination file path must have the '.qmd' extension.")
        sys.exit(1)

    with resources.open_text(templates, "template.qmd") as file:
        raw_template = file.read()

    rendered_template = Template(raw_template).render(
        filepath=destination,
        cms_admin_url=config.publish_base_url + config.publish_admin_path,
        project_folder=destination.parent.name,
        filename=destination.name,
    )
    destination.write_text(rendered_template)
    print(
        f"Successfully created template file {destination}.\n",
        "Open the file for further instructions.",
    )


def _preview(file_path: str, config: Config) -> None:
    if Path(file_path).suffix != ".qmd":
        print("Only Quarto Markdown (.qmd) files are supported.")
        sys.exit(1)

    publish_client = get_publish_client(config)

    _sync_updated_file(file_path, publish_client)

    print("Watching for file changes...")
    for changes in watch(file_path):
        print("Found changes:")
        print(changes)
        _sync_updated_file(file_path, publish_client)

def _sync_updated_file(file_path: str, publish_client: PublishClient) -> None:
    print("Syncing updated document...")
    preview_url = _sync_quarto_file(file_path, publish_client)
    print(f"Content synced successfully. Preview URL: {preview_url}")

def _sync_quarto_file(file_path: str, publish_client: PublishClient) -> str:
    pandoc_document = _quarto_to_pandoc(file_path)
    adapters = (
        PandocDocumentProcessor(),
        MimirContentParser(),
        LocalFileStorage(project_folder=Path(file_path).parent),
        publish_client,
    )
    return sync_document(pandoc_document, *adapters)


def _quarto_to_pandoc(file_path: str) -> str:
    try:
        result = subprocess.run(
            [
                "quarto",
                "render",
                file_path,
                "--to",
                "json",
                "-M",
                "include:false",
                "--output",
                "-",
            ],
            text=True,
            capture_output=True,
            check=True,
        )
        return result.stdout
    except subprocess.CalledProcessError as err:
        message = err.stderr.strip() or "Quarto failed with no error output"
        raise RuntimeError(f"Quarto render failed:\n{message}") from err



"""App configuration through environment variables."""

import os
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Config:
    publish_base_url: str
    publish_admin_path: str
    use_dapla_token_client: bool


_TEST_CONFIG = Config(
    publish_base_url="https://i.test.ssb.no",
    publish_admin_path="/xp/admin",
    use_dapla_token_client=True,
)

_LOCAL_CONFIG = Config(
    publish_base_url="http://localhost:8080",
    publish_admin_path="/admin",
    use_dapla_token_client=False,
)


def get_config(metadata_file_path: Path | None = None) -> Config:
    if os.environ.get("SSB_PUBMD_ENVIRONMENT") == "local":
        config = _LOCAL_CONFIG
    else:
        config = _TEST_CONFIG

    if os.environ.get("SSB_PUBMD_TOKEN_CLIENT") == "local":
        config.use_dapla_token_client = False

    return config